# -*- coding: utf-8 -*-


class MapLayeringRecord(object):
    """
    Represents a map layering entry with the view service, layer index and layer opacity.

    Args:
        view_service (str): View service.
        layer_index (str): Index for sorting the layering of the view services within a theme.
        layer_opacity (str): Opacity of a view service.
    """
    def __init__(self, view_service, layer_index, layer_opacity):

        self.view_service = view_service
        self.layer_index = layer_index
        self.layer_opacity = layer_opacity



# -*- coding: utf-8 -*-
import warnings


class OfficeRecord(object):
    """
    Responsible office record.

    Args:
        name (dict of unicode):  The official name of the authority (multilingual)
        uid (unicode): The unique identifier of the authority in the federal register
        office_at_web (dict of unicode): The URL to the office's homepage (multilingual)
        line1 (unicode): Complementary address information
        line2 (unicode): Complementary address information
        street (unicode): The street where the authority is located
        number (unicode): House number
        postal_code (integer): ZIP Code of the
        city (unicode): The city where the authority is located
            identifier (str): The identifier of the theme which might be used for linking to other elements.
    """
    def __init__(self, name, uid=None, office_at_web=None, line1=None, line2=None,
                 street=None, number=None, postal_code=None, city=None, identifier=None):
        if not isinstance(name, dict):
            warnings.warn('Type of "name" should be "dict"')

        self.name = name
        self.uid = uid
        self.office_at_web = office_at_web
        self.line1 = line1
        self.line2 = line2
        self.street = street
        self.number = number
        self.postal_code = postal_code
        self.city = city
        self.identifier = identifier



# -*- coding: utf-8 -*-
"""
Regarding to the main documentation you find here all the stuff which is related to the records. You can
imagine these record things as the in-application-currency. They are achieved and structured in the
:ref:`base-sources`. They are gathered by the :ref:`readers`. And they are combined, sorted, dismissed and
reorganized by the processor.

.. note:: This layer is not meant to be adapted in normal.
"""


class BaseLookUpRecord(object):
    """
    Translation of codes delivered by fed in the
    https://models.geo.admin.ch/V_D/OeREB/OeREBKRM_V2_0_Texte_20210714.xml

    The specification wants the extract codes translated to english, here we have the key as
    the value defined in the xml and the value is the corresponding translation.
    """
    lookup = {}

    def __init__(self, code):
        if code not in self.lookup.keys():
            raise AttributeError('wrong code for law status was deliverd. Only {0} are'
                                 ' allowed. Value was {1}'.format(self.lookup.keys(), code))
        self.code = code

    def get_extract_code(self):
        return self.lookup[self.code]



# -*- coding: utf-8 -*-


class MunicipalityRecord(object):
    """
    The base document class.

    Args:
        fosnr (int): The unique id bfs of the municipality.
        name (unicode): The zipcode for this address.
        published (bool): Is this municipality ready for publishing via server.
        logo (pyramid_oereb.lib.records.image.ImageRecord): The municipality logo.
        geom (str or None): The geometry which is representing this municipality as a WKT.
    """
    def __init__(self, fosnr, name, published, geom=None):
        self.fosnr = fosnr
        self.name = name
        self.published = published
        self.geom = geom



# -*- coding: utf-8 -*-
from pyramid_oereb.lib.sources import Base
from pyramid_oereb.lib.records.theme_document import ThemeDocumentRecord


class ThemeDocumentBaseSource(Base):
    """
    Base class for theme document sources.

    Attributes:
        records (list of pyramid_oereb.lib.records.theme.ThemeDocumentRecord): List of theme document records.
    """
    _record_class_ = ThemeDocumentRecord

    def read(self):
        """
        Every theme document source has to implement a read method. This method must accept no parameters.
        Because it should deliver all items available.
        If you want adapt to your own source for theme documents, this is the point where to hook in.
        """
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
from pyramid_oereb.lib.sources import BaseDatabaseSource
from pyramid_oereb.lib.sources.theme import ThemeBaseSource


class DatabaseSource(BaseDatabaseSource, ThemeBaseSource):

    def read(self):
        """
        Central method to read all theme entries.
        """
        session = self._adapter_.get_session(self._key_)
        try:
            results = session.query(self._model_).all()

            self.records = list()
            for result in results:
                self.records.append(self._record_class_(
                    result.code,
                    result.title,
                    result.extract_index,
                    result.sub_code,
                    identifier=result.id
                ))
        finally:
            session.close()



# -*- coding: utf-8 -*-
from pyramid_oereb.lib.sources import Base
from pyramid_oereb.lib.records.general_information import GeneralInformationRecord


class GeneralInformationBaseSource(Base):
    """
    Base class for GeneralInformation values source.

    Attributes:
        records (list of pyramid_oereb.lib.records.general_information.GeneralInformationRecord): List of
        general information records.
    """
    _record_class_ = GeneralInformationRecord

    def read(self):
        """
        Every general information source has to implement a read method.
        This method must accept no parameters. Because it should deliver all items available.
        If you want adapt to your own source for general information labels, this is the point
        where to hook in.
        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
        """
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
import logging

from pyramid.httpexceptions import HTTPInternalServerError

log = logging.getLogger(__name__)


def get_multilingual_element(value, language, not_null=True):
    multilingual_value = value.get(language, None)

    if multilingual_value is None and not_null:
        msg = 'Default language "{language}" is not available in: \n {element}'\
            .format(language=language, element=value)
        log.error(msg)
        raise HTTPInternalServerError()

    return multilingual_value



# -*- coding: utf-8 -*-
from pyramid_oereb.lib.sources import Base
from pyramid_oereb.lib.records.disclaimer import DisclaimerRecord


class DisclaimerBaseSource(Base):
    """
    Base class for exclusion of liability sources.

    Attributes:
        records (list of pyramid_oereb.lib.records.disclaimer.DisclaimerRecord): List
            of exclusion of liability records.
    """
    _record_class_ = DisclaimerRecord

    def read(self, params):
        """
        Every disclaimer source has to implement a read method. This method must accept no
        parameters. Because it should deliver all items available.
        If you want adapt to your own source for disclaimer, this is the point where to hook in.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
        """
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
import warnings


class GeneralInformationRecord(object):
    """
    Represents a general information entry with its title and content.
    Args:
        title (dict of unicode): The title of the information (multilingual)
        content (dict of unicode): The actual information (multilingual)
    """
    def __init__(self, title, content):

        if not isinstance(title, dict):
            warnings.warn('Type of "title" should be "dict"')
        if not isinstance(content, dict):
            warnings.warn('Type of "content" should be "dict"')

        self.title = title
        self.content = content

    def __str__(self):
        return '<{} -- title: {} content: {}>'.format(
            self.__class__.__name__, self.title, self.content)



# -*- coding: utf-8 -*-
from pyramid_oereb.lib.sources import Base
from pyramid_oereb.lib.records.documents import DocumentRecord


class DocumentBaseSource(Base):
    """
    Base class for document sources.

    Attributes:
        records (list of pyramid_oereb.lib.records.documents.DocumentRecord): List of document records.
    """
    _record_class_ = DocumentRecord

    def read(self, office_records):
        """
        Every document source has to implement a read method. This method must accept no parameters. Because
        it should deliver all items available.
        If you want adapt to your own source for documents, this is the point where to hook in.
        """
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
import warnings


class RealEstateTypeRecord(object):
    """
    Represents a real estate type entry with the code and it's display name.
    Args:
        code (str of unicode): The code for the real estate type.
        title (dict of unicode): The label title for the real estate type (multilingual).
    """
    def __init__(self, code, title):

        if not isinstance(code, str):
            warnings.warn('Type of "code" should be "str"')
        if not isinstance(title, dict):
            warnings.warn('Type of "title" should be "dict"')
        self.code = code
        self.title = title



# -*- coding: utf-8 -*-
"""
This is representing the applications working data schema. It provides buckets for general data you will need
to run this application in standard mode.
The buckets are:

* Municipality
* Real Estate
* Address
* Glossary
* Exclusion of liability

The geographical projection system which is used out of the box is LV95 aka EPSG:2056. Of course you can
configure a different one.
The name of the schema will be::

    pyramid_oereb_main

But you can change it also via configuration.

    .. note::

        Whenever you configure your own sqlalchemy ORM's to use them in this application you must imitate
        the behaviour of the ORM's here. This means the names class variables as well as the types of these
        variables.

"""
from pyramid_oereb.standard.models import get_office, get_document
from sqlalchemy import Column, PrimaryKeyConstraint, ForeignKey, UniqueConstraint
from sqlalchemy import Unicode, String, text, Integer, Boolean, Float
from geoalchemy2 import Geometry
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy_utils import JSONType
from sqlalchemy.orm import relationship

from pyramid_oereb.lib.config import Config

Base = declarative_base()
app_schema_name = Config.get('app_schema').get('name')
srid = Config.get('srid')


class RealEstate(Base):
    """
    The container where you can throw in all the real estates this application should have access to, for
    creating extracts.

    Attributes:
        id (int): The identifier. This is used in the database only and must not be set manually. If
            you  don't like it - don't care about.
        identdn (str): The identifier on cantonal level.
        number (str): The identifier on municipality level.
        egrid (str): The identifier on federal level (the all unique one...)
        type (str): The type of the real estate (This must base on DM01)
        canton (str): Which canton this real estate is situated in (use official shortened Version
            here. e.g. 'BE')
        municipality (str): The name of the municipality this real estate is situated in.
        subunit_of_land_register (str): The name of the maybe existing sub unit of land register if
            municipality in  combination with number does not offer a unique constraint.
            Else you can skip that.
        subunit_of_land_register_designation (str): The maybe existing cantonal description
            of the subunit_of_land_register e.g. «Grundbuchkreis», «Sektion», «Fraktion» etc.
        fosnr (int): The identifier of the municipality. It is the commonly known id_bfs.
        metadata_of_geographical_base_data (str): A link to the metadata which this geometry is
            based on which is  delivering a machine readable response format (XML).
        land_registry_area (str): The amount of the area of this real estate as it is declared in
            the land  registers information.
        limit (geoalchemy2.types.Geometry): The geometry of real estates border. For type
            information see geoalchemy2_.  .. _geoalchemy2:
            https://geoalchemy-2.readthedocs.io/en/0.2.4/types.html  docs dependent on the
            configured type.  This concrete one is MULTIPOLYGON
    """
    __table_args__ = {'schema': app_schema_name}
    __tablename__ = 'real_estate'
    id = Column(Integer, primary_key=True)
    identdn = Column(String, nullable=True)
    number = Column(String, nullable=True)
    egrid = Column(String, nullable=True)
    type = Column(String, nullable=False)
    canton = Column(String, nullable=False)
    municipality = Column(String, nullable=False)
    subunit_of_land_register = Column(String, nullable=True)
    subunit_of_land_register_designation = Column(String, nullable=True)
    fosnr = Column(Integer, nullable=False)
    metadata_of_geographical_base_data = Column(String, nullable=True)
    land_registry_area = Column(Integer, nullable=False)
    limit = Column(Geometry('MULTIPOLYGON', srid=srid))


class Municipality(Base):
    """
    The municipality is the place where you hold the information about all the municipalities
    you are having in your canton. This is used also in the applications process to check whether
    a municipality is published or not.

    Attributes:
        fosnr (int): The identifier of the municipality. It is the commonly known id_bfs or as or
            nofs in the  french part.
        name (str): The Name of the municipality.
        published (bool): Switch whether a municipality is published or not. This has direct
            influence on extract  generation.
        geom (geoalchemy2.types.Geometry): The geometry of municipality borders. For type
            information see geoalchemy2_.  .. _geoalchemy2:
            https://geoalchemy-2.readthedocs.io/en/0.2.4/types.html  docs dependent on the
            configured type.  This concrete one is MULTIPOLYGON
    """
    __table_args__ = {'schema': app_schema_name}
    __tablename__ = 'municipality'
    fosnr = Column(Integer, primary_key=True, autoincrement=False)
    name = Column(String, nullable=False)
    published = Column(Boolean, nullable=False, default=False, server_default=text('FALSE'))
    geom = Column(Geometry('MULTIPOLYGON', srid=srid), nullable=True)


class Address(Base):
    """
    The bucket you can throw all addresses in the application should be able to use for the get egrid
    webservice. This is a bypass for the moment. In the end it seems ways more flexible to bind
    a service here but if you like you can use it.

    Attributes:
        street_name (unicode): The street name for this address.
        street_number (str): The house number of this address.
        zip_code (int): The ZIP code for this address.
        geom (geoalchemy2.types.Geometry): The geometry of real estates border. For type information
            see geoalchemy2_.  .. _geoalchemy2:
            https://geoalchemy-2.readthedocs.io/en/0.2.4/types.html  docs dependent on the
            configured type.  This concrete one is POINT
    """
    __table_args__ = (
        PrimaryKeyConstraint("street_name", "street_number", "zip_code"),
        {'schema': app_schema_name}
    )
    __tablename__ = 'address'
    street_name = Column(Unicode, nullable=False)
    street_number = Column(String, nullable=False)
    zip_code = Column(Integer, nullable=False, autoincrement=False)
    geom = Column(Geometry('POINT', srid=srid))


Office = get_office(Base, app_schema_name, String)
Document = get_document(Base, app_schema_name, String, Office)


class Theme(Base):
    """
    The OEREB themes of the application

    Attributes:
        id (int): identifier, used in the database only
        code (str): OEREB code of the theme - unique and used to link each PublicLawRestriction
            with the corresponding theme
        sub_code (str): OEREB sub_code of the sub-theme: only available for sub themes.
        title (dict): the title of the theme as a multilingual dictionary
        extract_index (int): index to sort the themes in the extract, defined in the specification
    """
    __table_args__ = {'schema': app_schema_name}
    __tablename__ = 'theme'
    id = Column(String, primary_key=True)
    code = Column(String, nullable=False)
    sub_code = Column(String, nullable=True)
    title = Column(JSONType, nullable=False)
    extract_index = Column(Integer, nullable=False)
    UniqueConstraint(code, sub_code)


class Logo(Base):
    """
    The container for all logos and municipality coat of arms

    Attributes:
        code (str): The identifier given by a code
        logo (dict): The image encoded in base64
    """
    __table_args__ = {'schema': app_schema_name}
    __tablename__ = 'logo'
    id = Column(String, primary_key=True)
    code = Column(String, nullable=False)
    logo = Column(JSONType, nullable=False)


class RealEstateType(Base):
    """
    The container where you can throw in all the real estates type texts this application
    should have access to, for creating extracts.

    Attributes:
        code (str): The identifier on federal level.
        text (str): The text for the multilingual text.
    """
    __table_args__ = {'schema': app_schema_name}
    __tablename__ = 'real_estate_type'
    id = Column(String, primary_key=True)
    code = Column(String, nullable=False)
    title = Column(JSONType, nullable=False)


class Glossary(Base):
    """
    The bucket you can throw all items you want to have in the extracts glossary as reading help.

    Attributes:
        title (str): The title or abbreviation of a glossary item.
        content (str): The description or definition of a glossary item.
    """
    __table_args__ = {'schema': app_schema_name}
    __tablename__ = 'glossary'
    id = Column(String, primary_key=True)
    title = Column(JSONType, nullable=False)
    content = Column(JSONType, nullable=False)


class Disclaimer(Base):
    """
    The bucket you can throw all disclaimers in the application should be able to use.

    Attributes:
        title (str): The title which the disclaimer item has.
        content (str): The content which the disclaimer item has.
    """
    __table_args__ = {'schema': app_schema_name}
    __tablename__ = 'disclaimer'
    id = Column(String, primary_key=True)
    title = Column(JSONType, nullable=False)
    content = Column(JSONType, nullable=False)


class LawStatus(Base):
    """
    The container where you can throw in all the law status texts this application
    should have access to, for creating extracts.
    Attributes:
        code (str): The identifier on federal level.
        title (JSONType): The text for the multilingual text.
    """
    __table_args__ = {'schema': app_schema_name}
    __tablename__ = 'law_status'
    code = Column(String, primary_key=True)
    title = Column(JSONType, nullable=False)


class DocumentTypeText(Base):
    """
    The element holding the different document types and their translations.

    Attributes:
        code (str): The identifier given by a code
        title (str): The display name for the document type
    """
    __table_args__ = {'schema': app_schema_name}
    __tablename__ = 'document_types'
    code = Column(String, primary_key=True)
    title = Column(JSONType, nullable=False)


class GeneralInformation(Base):
    """
    The bucket to store the general information about the OEREB cadastre

    Attributes:
        title (dict): The title of the general information (multilingual)
        content (dict): The actual information (multilingual)
    """
    __table_args__ = {'schema': app_schema_name}
    __tablename__ = 'general_information'
    id = Column(String, primary_key=True)
    title = Column(JSONType, nullable=False)
    content = Column(JSONType, nullable=False)


class ThemeDocument(Base):
    """
    Meta bucket (join table) for the relationship between theme and documents.
    Attributes:
        theme_id (str): The foreign key to the theme which has
            relation to  a document.
        document_id (str): The foreign key to the document which has relation to the public law
            restriction.
        theme (Theme):
            The dedicated relation to the theme instance from database.
        document (Document):
            The dedicated relation to the document instance from database.
        article_numbers (dict): relevant articles for the theme document relation
    """
    __tablename__ = 'theme_document'
    __table_args__ = (
        PrimaryKeyConstraint("theme_id", "document_id"),
        {'schema': app_schema_name}
    )

    theme_id = Column(
        ForeignKey(Theme.id),
        nullable=False
    )
    document_id = Column(
        ForeignKey(Document.id),
        nullable=False
    )
    theme = relationship(
        Theme,
        backref='legal_provisions'
    )
    document = relationship(
        Document
    )
    article_numbers = Column(JSONType, nullable=True)


class MapLayering(Base):
    """
    Attributes:
        view_service (dict): Darstellungsdienst
        layer_index (int): Index for sorting the layering of the view services for a theme
        layer_opacity (float): Opacity of a view service
    """
    __table_args__ = {'schema': app_schema_name}
    __tablename__ = 'map_layering'
    id = Column(String, primary_key=True)
    view_service = Column(JSONType, nullable=False)
    layer_index = Column(Integer, nullable=False)
    layer_opacity = Column(Float, nullable=False)



# -*- coding: utf-8 -*-
from pyramid_oereb.standard.xtf_import.util import parse_integer, parse_string, parse_multilingual_text, \
    parse_ref


class Document(object):

    TAG_TYPE = 'Typ'
    TAG_TITLE = 'Titel'
    TAG_ABBREVIATION = 'Abkuerzung'
    TAG_OFFICIAL_NUMBER = 'OffizielleNr'
    TAG_ONLY_IN_MUNICIPALITY = 'NurInGemeinde'
    TAG_TEXT_AT_WEB = 'TextImWeb'
    TAG_FILE = 'Dokument'
    TAG_INDEX = 'AuszugIndex'
    TAG_LAW_STATUS = 'Rechtsstatus'
    TAG_PUBLISHED_FROM = 'publiziertAb'
    TAG_PUBLISHED_UNTIL = 'publiziertBis'
    TAG_RESPONSIBLE_OFFICE = 'ZustaendigeStelle'

    def __init__(self, session, model):
        self._session = session
        self._model = model

    def parse(self, document):  # pragma: no cover
        instance = self._model(
            id=document.attrib['TID'],
            document_type=parse_string(document, self.TAG_TYPE),
            title=parse_multilingual_text(document, self.TAG_TITLE),
            abbreviation=parse_multilingual_text(document, self.TAG_ABBREVIATION),
            official_number=parse_multilingual_text(document, self.TAG_OFFICIAL_NUMBER),
            only_in_municipality=parse_integer(document, self.TAG_ONLY_IN_MUNICIPALITY),
            text_at_web=parse_multilingual_text(document, self.TAG_TEXT_AT_WEB),
            index=parse_integer(document, self.TAG_INDEX),
            law_status=parse_string(document, self.TAG_LAW_STATUS),
            published_from=parse_string(document, self.TAG_PUBLISHED_FROM),
            published_until=parse_string(document, self.TAG_PUBLISHED_UNTIL),
            office_id=parse_ref(document, self.TAG_RESPONSIBLE_OFFICE)
        )
        self._session.add(instance)



# -*- coding: utf-8 -*-
import optparse

from pyramid_oereb.lib.config import Config
from pyramid_oereb.standard.sources.plr import StandardThemeConfigParser
from pyramid_oereb.standard import drop_sql, execute_sql


def drop_tables_from_standard_configuration(configuration_yaml_path, section='pyramid_oereb'):
    """
    Drops all schemas which are defined in the passed yaml file: <section>.<plrs>.[<plr>.<code>]. The code
    must be camel case. It will be transformed to snake case and used as schema name.
    Drops all tables inside the created schemas.

    Args:
        configuration_yaml_path (str): The absolute path to the yaml file which contains the plr
            definitions.
        section (str): The section in yaml file where the plrs are configured in. Default is 'pyramid_oereb'.
    """
    if Config.get_config() is None:
        Config.init(configuration_yaml_path, section)

    main_db_connection = Config.get('app_schema').get('db_connection')
    main_schema_name = Config.get('app_schema').get('name')
    sql = drop_sql(main_schema_name)
    execute_sql(main_db_connection, sql)

    for theme_config in Config.get('plrs'):
        if theme_config.get('standard'):
            config_parser = StandardThemeConfigParser(**theme_config)
            models = config_parser.get_models()
            theme_db_connection = models.db_connection
            theme_schema_name = models.schema_name
            sql = drop_sql(theme_schema_name)
            execute_sql(theme_db_connection, sql)


def drop_standard_tables():
    parser = optparse.OptionParser(
        usage='usage: %prog [options]',
        description='Create all content for the standard database'
    )
    parser.add_option(
        '-c', '--configuration',
        dest='configuration',
        metavar='YAML',
        type='string',
        help='The absolute path to the configuration yaml file (standard is: pyramid_oereb.yml).'
    )
    parser.add_option(
        '-s', '--section',
        dest='section',
        metavar='SECTION',
        type='string',
        default='pyramid_oereb',
        help='The section which contains configruation (default is: pyramid_oereb).'
    )
    options, args = parser.parse_args()
    if not options.configuration:
        parser.error('No configuration file set.')
    drop_tables_from_standard_configuration(
        configuration_yaml_path=options.configuration,
        section=options.section
    )



# -*- coding: utf-8 -*-
import warnings


class ThemeRecord(object):
    """Creates a new theme record."""

    def __init__(self, code, title, extract_index, sub_code=None, document_records=None, identifier=None):
        """
        Args:
            code (unicode): The theme's code.
            title (dict of unicode): The multilingual description.
            extract_index (int): Index to sort themes in the extract.
            sub_code (unicode): The code of the sub_theme. Is none for themes.
            document_records (list): Documents that relates to this theme
            identifier (str): The identifier of the theme which might be used for linking to other elements.
        """
        if not isinstance(title, dict):
            warnings.warn('Type of "title" should be "dict"')

        self.code = code
        self.sub_code = sub_code
        self.title = title
        self.extract_index = extract_index
        self.document_records = document_records
        self.identifier = identifier

    def __str__(self):
        return '<{} -- code: {} sub code: {} title: {} extract index: {} document records: {}>'.format(
            self.__class__.__name__, self.code, self.sub_code, self.title,
            self.extract_index, self.document_records)



# -*- coding: utf-8 -*-
from pyramid_oereb.lib.sources import BaseDatabaseSource
from pyramid_oereb.lib.sources.document import DocumentBaseSource
from pyramid_oereb import Config


class DatabaseSource(BaseDatabaseSource, DocumentBaseSource):

    def read(self, office_records):
        """
        Central method to read all theme entries.
        """
        session = self._adapter_.get_session(self._key_)
        try:
            results = session.query(self._model_).all()
            self.records = list()
            for result in results:
                office_record_match = None
                for office_record in office_records:
                    if office_record.identifier == result.office_id:
                        office_record_match = office_record_match
                        break
                self.records.append(
                    self._record_class_(
                        document_type=Config.get_document_type_by_code(
                            result.document_type
                        ),
                        index=result.index,
                        law_status=Config.get_law_status_by_code(
                            result.law_status
                        ),
                        title=result.title,
                        responsible_office=office_record,
                        published_from=result.published_from,
                        published_until=result.published_until,
                        text_at_web=result.text_at_web,
                        abbreviation=result.abbreviation,
                        official_number=result.official_number,
                        only_in_municipality=result.only_in_municipality,
                        article_numbers=None,
                        file=result.file,
                        identifier=result.id
                    )
                )
        finally:
            session.close()



# -*- coding: utf-8 -*-

def from_multilingual_text_to_dict(de=None, fr=None, it=None, rm=None, en=None):
    if de is None and fr is None and it is None and rm is None and en is None:
        return None
    else:
        dict_var = dict()
        if de is not None:
            dict_var['de'] = de
        if fr is not None:
            dict_var['fr'] = fr
        if it is not None:
            dict_var['it'] = it
        if rm is not None:
            dict_var['rm'] = rm
        if en is not None:
            dict_var['en'] = en
        return dict_var


def from_multilingual_uri_to_dict(obj):
    """
    obj (pyramid_oereb.contrib.models.interlis_2_3.MultilingualUri)
    """
    for ml_uris in obj:
        var_dict = dict()
        for row in ml_uris.localised_uri:
            var_dict[row.language] = row.text
        return var_dict


def from_multilingual_blob_to_dict(obj):
    """
    obj (pyramid_oereb.contrib.models.interlis_2_3.MultilingualBlob)
    """
    for ml_blobs in obj:
        var_dict = dict()
        for row in ml_blobs.localised_blob:
            var_dict[row.language] = row.blob
        return var_dict



# -*- coding: utf-8 -*-

from geoalchemy2.elements import _SpatialElement

from pyramid_oereb.lib.sources import BaseDatabaseSource
from geoalchemy2.shape import to_shape

from pyramid_oereb.lib.sources.municipality import MunicipalityBaseSource


class DatabaseSource(BaseDatabaseSource, MunicipalityBaseSource):

    def read(self, params, fosnr=None):
        """
        Central method to read a municipality by it's id_bfs identifier.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
            fosnr (int or None): The federal number of the municipality defined by the statistics office.
        """
        session = self._adapter_.get_session(self._key_)
        try:
            self.records = list()
            if fosnr:
                results = session.query(self._model_).filter(self._model_.fosnr == fosnr).all()
            else:
                results = session.query(self._model_).all()
            for result in results:
                self.records.append(self._record_class_(
                    result.fosnr,
                    result.name,
                    result.published,
                    geom=to_shape(result.geom).wkt if isinstance(
                        result.geom, _SpatialElement) else None,
                ))
        finally:
            session.close()



# -*- coding: utf-8 -*-


class LawStatusRecord(object):
    """
    Law status record.
    """

    def __init__(self, code, title):
        """
        Create a new law status record.

        Args:
            code (str or unicode): The code of the law status. It must be "inKraft" or
            "AenderungMitVorwirkung" or "AenderungOhneVorwirkung" every other value will
            raise an error.
        title (dict of unicode): The multilingual law status description.
        """
        self.code = code
        self.title = title



# -*- coding: utf-8 -*-
import warnings
from datetime import datetime
import uuid


class ExtractRecord(object):
    """The extract base class."""

    creation_date = None
    """datetime.datetime: The date and time of the extract creation."""

    electronic_signature = None
    """unicode or None: Digital signature for the extract."""

    concerned_theme = None
    """list of pyramid_oereb.lib.records.theme.ThemeRecord: List of concerned themes."""

    not_concerned_theme = None
    """list of pyramid_oereb.lib.records.theme.ThemeRecord: List of not concerned themes."""

    theme_without_data = None
    """list of pyramid_oereb.lib.records.theme.ThemeRecord: List of themes without data."""

    extract_identifier = None
    """unicode: The extract identifier (UUID)."""

    qr_code = None
    """binary or None: QR code for the extract as binary string."""

    def __init__(self, real_estate, logo_plr_cadastre, federal_logo, cantonal_logo, municipality_logo,
                 plr_cadastre_authority, update_date_os, disclaimers=None, glossaries=None,
                 concerned_theme=None, not_concerned_theme=None, theme_without_data=None,
                 general_information=None):
        """
        Args:
            real_estate (pyramid_oereb.lib.records.real_estate.RealEstateRecord): The real estate in its
                record representation.
            logo_plr_cadastre (pyramid_oereb.lib.records.logo.LogoRecord): Image file of the PLR logo.
            federal_logo (pyramid_oereb.lib.records.image.ImageRecord):Image file of the federal logo.
            cantonal_logo (pyramid_oereb.lib.records.image.ImageRecord): Image file of the cantonal logo.
            municipality_logo (pyramid_oereb.lib.records.image.ImageRecord): Image file of the municipality
                logo.
            plr_cadastre_authority (pyramid_oereb.lib.records.office.OfficeRecord): The authority which is
                responsible for the PLR cadastre.
            update_date_os (datetime): Last update of the official survey used as base map in the extract.
            disclaimer (list of
                pyramid_oereb.lib.records.disclaimer.DisclaimerRecord or None): Disclaimers for the extract.
            glossaries (list of pyramid_oereb.lib.records.glossary.GlossaryRecord): Glossaries for the
                extract.
            concerned_theme (list of pyramid_oereb.lib.records.theme.ThemeRecord or None): Concerned themes.
            not_concerned_theme (list of pyramid_oereb.lib.records.theme.ThemeRecord or None): Not concerned
                themes.
            theme_without_data (list of pyramid_oereb.lib.records.theme.ThemeRecord or None): Themes without
                data.
            general_information (list of dict): General information for the static extract as multilingual
                text.
        """
        if not isinstance(update_date_os, datetime):
            warnings.warn('Type of "update_date_os" should be "datetime.datetime"')
        if general_information and not isinstance(general_information, list):
            warnings.warn('Type of "general_information" should be "list"')

        self.update_date_os = update_date_os
        self.general_information = general_information
        self.extract_identifier = str(uuid.uuid4())
        self.real_estate = real_estate
        if concerned_theme:
            self.concerned_theme = concerned_theme
        else:
            self.concerned_theme = []
        if not_concerned_theme:
            self.not_concerned_theme = not_concerned_theme
        else:
            self.not_concerned_theme = []
        if theme_without_data:
            self.theme_without_data = theme_without_data
        else:
            self.theme_without_data = []
        self.creation_date = datetime.now()
        self.logo_plr_cadastre = logo_plr_cadastre
        self.federal_logo = federal_logo
        self.cantonal_logo = cantonal_logo
        self.municipality_logo = municipality_logo
        self.plr_cadastre_authority = plr_cadastre_authority
        if disclaimers:
            self.disclaimers = disclaimers
        else:
            self.disclaimers = []
        if glossaries:
            self.glossaries = glossaries
        else:
            self.glossaries = []



# -*- coding: utf-8 -*-
import math
import os
import shutil
import zipfile
import re
from datetime import datetime
from uuid import uuid4

import requests
import logging

from defusedxml.ElementTree import parse
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from tqdm import tqdm

from pyramid_oereb.lib.config import Config
from pyramid_oereb.standard.sources.plr import StandardThemeConfigParser
from pyramid_oereb.standard.xtf_import.base_refinement import BaseRefinement
from pyramid_oereb.standard.xtf_import.document import Document
from pyramid_oereb.standard.xtf_import.geometry import Geometry
from pyramid_oereb.standard.xtf_import.legend_entry import LegendEntry
from pyramid_oereb.standard.xtf_import.office import Office
from pyramid_oereb.standard.xtf_import.public_law_restriction import PublicLawRestriction
from pyramid_oereb.standard.xtf_import.public_law_restriction_document import PublicLawRestrictionDocument
from pyramid_oereb.standard.xtf_import.util import get_tag
from pyramid_oereb.standard.xtf_import.view_service import ViewService


class FederalTopic(object):
    """
    Download and import data for a specified federal topic.
    """

    TAG_DATASECTION = 'DATASECTION'
    TAG_TRANSFER_STRUCTURE = 'OeREBKRMtrsfr_V1_1.Transferstruktur'
    TAG_REFERENCED_LAWS = 'OeREBKRMvs_V1_1.HinweiseGesetzlicheGrundlagen'
    TAG_OFFICE = 'OeREBKRMvs_V1_1.Vorschriften.Amt'
    TAG_DOCUMENT = 'OeREBKRMvs_V1_1.Vorschriften.Dokument'
    TAG_VIEW_SERVICE = 'OeREBKRMtrsfr_V1_1.Transferstruktur.DarstellungsDienst'
    TAG_PLR = 'OeREBKRMtrsfr_V1_1.Transferstruktur.Eigentumsbeschraenkung'
    TAG_GEOMETRY = 'OeREBKRMtrsfr_V1_1.Transferstruktur.Geometrie'
    TAG_PUBLIC_LAW_RESTRICTION_DOCUMENT = 'OeREBKRMtrsfr_V1_1.Transferstruktur.HinweisVorschrift'
    TAG_BASE_REFINEMENT = 'OeREBKRMtrsfr_V1_1.Transferstruktur.GrundlageVerfeinerung'

    def __init__(self, configuration_file, topic_code, section='pyramid_oereb',
                 c2ctemplate_style=False, arc_max_diff=0.001, arc_precision=3, tmp_dir='.',
                 srid=None):
        """

        Args:
            configuration_file (str): Path to the configuration file to be used.
            topic_code (str): The code of the federal topic to be updated.
            section (str): The section within the configuration file. (default: 'pyramid_oereb')
            c2ctemplate_style (bool): True if the yaml use a c2c template style (vars.[section]).
                Default is False.
            arc_max_diff (float): Maximum difference between arc and line segment for stroking.
                (default: 0.001)
            arc_precision (int): Coordinate precision for generated arc points. (default: 3)
            tmp_dir (str): Directory used as temporary working directory. (default: '.')
        """
        self._log = logging.getLogger('import_federal_topic')
        Config.init(configuration_file, section, c2ctemplate_style)
        self._settings = Config.get_config()

        topic_settings = None
        for topic in self._settings.get('plrs'):
            if topic.get('code') == topic_code:
                topic_settings = topic
        if topic_settings is None:
            self._log.error('Cannot find topic {0} in {1}'.format(topic_code, configuration_file))
            exit(1)
        self._srid = srid or self._settings.get('srid')
        if self._srid is None:
            self._log.error('No SRID defined in configuration or passed as argument')
            exit(1)
        self._tmp_dir = tmp_dir
        self._arc_max_diff = arc_max_diff
        self._arc_precision = arc_precision
        self._topic_settings = topic_settings
        self._connection = topic_settings.get('source').get('params').get('db_connection')
        config_parser = StandardThemeConfigParser(**topic_settings)
        self._models = config_parser.get_models()
        self._file_id = '{0}'.format(uuid4())
        self._checksum = None
        self._data_integration_office_id = None

    def load(self, xtf_files, force=False):
        """
        Updates the data for a certain federal topic.

        Args:
            xtf_files (list of str): Files to be parsed. The first one should be the XML file containing the
                federal laws, the second one the XTF file with the PLR data.
            force (bool): Set to `True` to ignore the result of the checksum comparison and update the data
                anyway.
        """

        engine = create_engine(self._connection)
        Session = sessionmaker(bind=engine)
        session = Session()

        try:
            if force or self._compare_checksum(session):
                self._log.info('Starting import of topic {0}'.format(self._topic_settings.get('code')))
                self._log.info('Using SRID: {0}'.format(self._srid))
                self._log.info('Maximum difference for arc points: {0}'.format(self._arc_max_diff))
                self._log.info('Arc point coordinate precision: {0}'.format(self._arc_precision))

                self._truncate_schema(session)
                topic_source = None
                for xtf_file in xtf_files:
                    if xtf_file.endswith('.xtf'):
                        topic_source = os.path.basename(xtf_file)
                    self._log.info('Parsing {0}'.format(xtf_file))
                    content = parse(xtf_file)
                    for element in content.getroot():
                        if get_tag(element) == self.TAG_DATASECTION:
                            self._parse_datasection(session, element)

                self._update_data_integration(session, topic_source)

                self._log.info('Committing import')
                session.commit()

                self._log.info('Finished import of topic {0}'.format(self._topic_settings.get('code')))

        except Exception:
            session.rollback()
            self._log.exception('An error occurred during the import:')
            self.cleanup_files()
            exit(1)

        finally:
            session.close()

    def _compare_checksum(self, session):
        """
        Compares the checksums within the downloaded data and the database. Returns `False` if the checksums
        match or no checksum is available within the downloaded data. If the checksums are different or no
        checksum is available in the database, `True` is returned.

        Args:
            session (sqlalchemy.orm.session.Session): The SQLAlchemy session for database interaction.

        Returns:
            bool: `True` if the data should be updated, `False` otherwise.
        """
        if self._checksum is None:
            self._log.info('No checksum available for comparison. Skipping import of data. Use "--force" to '
                           'import data anyway.')
            return False
        data_integration = session.query(self._models.DataIntegration).first()
        if data_integration is None:
            return True
        elif getattr(data_integration, 'checksum') == self._checksum:
            self._log.info('Checksums in data and database are identical. Skipping import.')
            return False
        else:
            self._log.info('Checksum in data has changed: {0} -> {1}'.format(
                getattr(data_integration, 'checksum'),
                self._checksum
            ))
            return True

    def _truncate_schema(self, session):
        """
        Truncates the tables for the specified topic.

        Args:
            session (sqlalchemy.orm.session.Session): The SQLAlchemy session for database interaction.
        """
        models = [
            self._models.Office,
            self._models.DataIntegration,
            self._models.Document,
            self._models.ViewService,
            self._models.LegendEntry,
            self._models.PublicLawRestriction,
            self._models.Geometry,
            self._models.PublicLawRestrictionBase,
            self._models.PublicLawRestrictionRefinement,
            self._models.PublicLawRestrictionDocument
        ]
        self._log.info('Truncating tables:')
        for model in models:
            self._log.info('    {0}.{1}'.format(model.__table__.schema, model.__table__.name))
            session.execute('TRUNCATE {schema}.{table} CASCADE;'.format(
                schema=model.__table__.schema,
                table=model.__table__.name
            ))

    def _update_data_integration(self, session, topic_source):
        """
        Update the table `data_integration` with the data of the current import.

        Args:
            session (sqlalchemy.orm.session.Session): The SQLAlchemy session for database interaction.
            topic_source (str): Name of the parsed XTF file.
        """
        instance = self._models.DataIntegration(
            id=topic_source,
            date=datetime.now().isoformat(),
            office_id=self._data_integration_office_id,
            checksum=self._checksum
        )
        session.add(instance)

    def _parse_datasection(self, session, datasection):
        """
        Parses the data section

        Args:
            session (sqlalchemy.orm.session.Session): The SQLAlchemy session for database interaction.
            datasection (lxml.etree.Element): The data section element.
        """
        for element in datasection:
            tag = get_tag(element)
            if tag in [self.TAG_TRANSFER_STRUCTURE, self.TAG_REFERENCED_LAWS]:
                laws = (tag == self.TAG_REFERENCED_LAWS)
                self._parse_transfer_structure(session, element, laws=laws)

    def _parse_transfer_structure(self, session, transfer_structure, laws=False):
        """
        Parses the transfer structure content.

        Args:
            session (sqlalchemy.orm.session.Session): The SQLAlchemy session for database interaction.
            transfer_structure (lxml.etree.Element): The transfer structure element.
            laws (bool): True if the parsed file is the XML containing the federal laws.
        """

        office = Office(session, self._models.Office)
        document = Document(session, self._models.Document)
        legend_entry = LegendEntry(session, self._models.LegendEntry, self._topic_settings.get('code'))
        view_service = ViewService(session, self._models.ViewService, legend_entry)
        public_law_restriction = PublicLawRestriction(
            session,
            self._models.PublicLawRestriction,
            self._topic_settings.get('code')
        )
        geometry = Geometry(
            session,
            self._models.Geometry,
            self._topic_settings.get('geometry_type'),
            self._srid,
            arc_max_diff=self._arc_max_diff,
            arc_precision=self._arc_precision
        )
        public_law_restriction_document = PublicLawRestrictionDocument(
            session,
            self._models.PublicLawRestrictionDocument
        )
        base_refinement = BaseRefinement(
            session,
            self._models.PublicLawRestrictionBase,
            self._models.PublicLawRestrictionRefinement
        )

        for element in transfer_structure:
            tag = get_tag(element)
            if tag == self.TAG_OFFICE:
                # Use the last office ID for data integration
                self._data_integration_office_id = element.attrib['TID']
                office.parse(element)
            elif tag == self.TAG_DOCUMENT:
                document.parse(element)
            elif tag == self.TAG_VIEW_SERVICE:
                view_service.parse(element)
            elif tag == self.TAG_PLR:
                public_law_restriction.parse(element)
            elif tag == self.TAG_GEOMETRY:
                geometry.parse(element)
            elif tag == self.TAG_PUBLIC_LAW_RESTRICTION_DOCUMENT:
                public_law_restriction_document.parse(element)
            elif tag == self.TAG_BASE_REFINEMENT:
                base_refinement.parse(element)
            else:
                self._log.error('NOT IMPLEMENTED: {0}'.format(get_tag(element)))

    def download_data(self):  # pragma: no cover
        """
        Downloads the federal data for the specified topic.
        """
        file_path = os.path.join(self._tmp_dir, '{0}.zip'.format(self._file_id))
        url = self._topic_settings.get('download')
        if url is None:
            self._log.error('Missing download URL in configuration')
            exit(1)
        self._log.info('Downloading data from {0}'.format(url))
        response = requests.get(url, stream=True, proxies=self._settings.get('proxies'))
        if response.status_code < 400:
            total_size = int(response.headers.get('content-length', 0))
            block_size = 1024
            wrote = 0
            self._log.info('Saving file to {0}'.format(file_path))
            with open(file_path, 'wb') as handle:
                for data in tqdm(
                        response.iter_content(block_size),
                        total=math.ceil(total_size//block_size),
                        unit='KB',
                        unit_scale=True
                ):
                    wrote = wrote + len(data)
                    handle.write(data)
        else:
            self._log.error('HTTP {0} - {1}'.format(response.status_code, response.content))

    def unzip_data(self):
        """
        Extracts the downloaded data into a temporary directory.
        """
        file_path = os.path.join(self._tmp_dir, '{0}.zip'.format(self._file_id))
        zip_path = os.path.join(self._tmp_dir, '{0}'.format(self._file_id))
        self._log.info('Extracting {0} to {1}'.format(file_path, zip_path))
        with open(file_path, 'rb') as f:
            zip_file = zipfile.ZipFile(f)
            for member in zip_file.namelist():
                if member.endswith('.xtf') or member.endswith('.md5.txt') \
                        or (member.endswith('.xml') and member.find('Gesetze') > -1):
                    self._log.info('\t{0}'.format(member))
                    zip_file.extract(member, zip_path)
        self._log.info('Deleting {0}'.format(file_path))
        os.remove(file_path)

    def collect_files(self):
        """
        Collects the extracted files for the data import.
        """
        path = os.path.join(self._tmp_dir, '{0}'.format(self._file_id))
        law_source = topic_source = None
        for item in os.listdir(path):
            candidate = os.path.join(path, item)
            if os.path.isfile(candidate):
                if item.endswith('.xtf'):
                    topic_source = candidate
                elif re.match(r'OeREBKRM_V.+_Gesetze_[0-9]{8}.xml', item):
                    law_source = candidate
        files = list()
        if law_source is not None:
            self._log.info('Found law source file: {0}'.format(law_source))
            files.append(law_source)
        if topic_source is not None:
            self._log.info('Found topic source file: {0}'.format(topic_source))
            files.append(topic_source)
        return files

    def cleanup_files(self):
        """
        Removes the temporary files and directory.
        """
        path = os.path.join(self._tmp_dir, '{0}'.format(self._file_id))
        zip_file = os.path.join(self._tmp_dir, '{0}.zip'.format(self._file_id))
        if os.path.isfile(zip_file):
            self._log.info('Deleting file {0}'.format(zip_file))
            os.remove(zip_file)
        if os.path.isdir(path):
            self._log.info('Deleting folder {0}'.format(path))
            shutil.rmtree(path)

    def read_checksum(self):
        """
        Reads the data's checksum from the extracted files.
        """
        path = os.path.join(self._tmp_dir, '{0}'.format(self._file_id))
        for item in os.listdir(path):
            candidate = os.path.join(path, item)
            if os.path.isfile(candidate):
                if candidate.endswith('.md5.txt'):
                    with open(candidate, 'r') as f:
                        self._checksum = '{0}'.format(f.read()).strip()
        self._log.info('Current checksum: {0}'.format(self._checksum))



# -*- coding: utf-8 -*-
import optparse
import logging

from pyramid.path import DottedNameResolver

from pyramid_oereb.lib.config import Config
from pyramid_oereb.standard.sources.plr import StandardThemeConfigParser
from pyramid_oereb.standard import tables, create_sql, create_tables_sql, execute_sql

logging.basicConfig()
log = logging.getLogger(__name__)


def create_theme_tables_(theme_config, tables_only=False, sql_file=None):
    """
    Create the tables for a specific theme.

    Args:
        theme_config (dict): The configuration section for the specific theme.
        tables_only (bool): True to skip creation of schema. Default is False.
        sql_file (file): The file to generate. Default is None (in the database).
    """
    if theme_config.get('standard'):
        config_parser = StandardThemeConfigParser(**theme_config)
        models = config_parser.get_models()
        theme_db_connection = models.db_connection
        theme_schema_name = models.schema_name
        theme_tables = tables(models.Base)
        if tables_only:
            sql = create_tables_sql(theme_db_connection, theme_tables)
        else:
            sql = create_sql(theme_schema_name, theme_db_connection, theme_tables)

        if sql_file is None:
            execute_sql(theme_db_connection, sql)
        else:
            sql_file.write(sql)


def create_tables_from_standard_configuration(
        configuration_yaml_path, section='pyramid_oereb', c2ctemplate_style=False, tables_only=False,
        sql_file=None):
    """
    Creates all schemas which are defined in the passed yaml file: <section>.<plrs>.[<plr>.<code>]. The code
    must be camel case. It will be transformed to snake case and used as schema name.
    Creates all tables inside the created schemas. This only affects the sqlalchemy models which are defined
    with the Base class from pyramid_oereb.standard.models.

    Args:
        configuration_yaml_path (str): The absolute path to the yaml file which contains the plr
            definitions.
        section (str): The section in yaml file where the plrs are configured in. Default is 'pyramid_oereb'.
        c2ctemplate_style (bool): True if the yaml use a c2c template style (vars.[section]).
            Default is False.
        tables_only (bool): True to skip creation of schema. Default is False.
        sql_file (file): the file to generate. Default is None (in the database).
    """
    if Config.get_config() is None:
        Config.init(configuration_yaml_path, section, c2ctemplate_style)

    main_base_class = DottedNameResolver().maybe_resolve('{package}.Base'.format(
        package=Config.get('app_schema').get('models')
    ))
    main_db_connection = Config.get('app_schema').get('db_connection')
    main_schema_name = Config.get('app_schema').get('name')
    main_tables = tables(main_base_class)
    if tables_only:
        sql = create_tables_sql(main_db_connection, main_tables)
    else:
        sql = create_sql(main_schema_name, main_db_connection, main_tables)
    if sql_file is None:
        execute_sql(main_db_connection, sql)
    else:
        sql_file.write(sql)

    for theme_config in Config.get('plrs'):
        if theme_config.get('standard'):
            create_theme_tables_(theme_config, tables_only=tables_only, sql_file=sql_file)


def create_standard_tables():
    parser = optparse.OptionParser(
        usage='usage: %prog [options]',
        description='Create all content for the standard database'
    )
    parser.add_option(
        '-c', '--configuration',
        dest='configuration',
        metavar='YAML',
        type='string',
        help='The absolute path to the configuration yaml file.'
    )
    parser.add_option(
        '-s', '--section',
        dest='section',
        metavar='SECTION',
        type='string',
        default='pyramid_oereb',
        help='The section which contains configuration (default is: pyramid_oereb).'
    )
    parser.add_option(
        '-T', '--tables-only',
        dest='tables_only',
        action='store_true',
        default=False,
        help='Use this flag to skip the creation of the schema.'
    )
    parser.add_option(
        '--sql-file',
        type='string',
        help='Generate an SQL file.'
    )
    parser.add_option(
        '--c2ctemplate-style',
        dest='c2ctemplate_style',
        action='store_true',
        default=False,
        help='Is the yaml file using a c2ctemplate style (starting with vars)'
    )
    options, args = parser.parse_args()
    if not options.configuration:
        parser.error('No configuration file set.')

    if options.sql_file is None:
        create_tables_from_standard_configuration(
            configuration_yaml_path=options.configuration,
            section=options.section,
            c2ctemplate_style=options.c2ctemplate_style,
            tables_only=options.tables_only
        )
    else:
        with open(options.sql_file, 'w') as sql_file:
            create_tables_from_standard_configuration(
                configuration_yaml_path=options.configuration,
                section=options.section,
                c2ctemplate_style=options.c2ctemplate_style,
                sql_file=sql_file
            )


def create_theme_tables():
    parser = optparse.OptionParser(
        usage='usage: %prog [options]',
        description='Create all content for the standard database'
    )
    parser.add_option(
        '-c', '--configuration',
        dest='configuration',
        metavar='YAML',
        type='string',
        help='The absolute path to the configuration yaml file.'
    )
    parser.add_option(
        '-s', '--section',
        dest='section',
        metavar='SECTION',
        type='string',
        default='pyramid_oereb',
        help='The section which contains configuration (default is: pyramid_oereb).'
    )
    parser.add_option(
        '-t', '--theme',
        dest='theme',
        metavar='CODE',
        type='string',
        help='The code of the theme to be created.'
    )
    parser.add_option(
        '-T', '--tables-only',
        dest='tables_only',
        action='store_true',
        default=False,
        help='Use this flag to skip the creation of the schema.'
    )
    parser.add_option(
        '--sql-file',
        type='string',
        help='Generate an SQL file.'
    )
    parser.add_option(
        '--c2ctemplate-style',
        dest='c2ctemplate_style',
        action='store_true',
        default=False,
        help='Is the yaml file using a c2ctemplate style (starting with vars)'
    )
    options, args = parser.parse_args()
    if not options.configuration:
        parser.error('No configuration file set.')
    if not options.theme:
        parser.error('No theme code specified.')

    if Config.get_config() is None:
        Config.init(
            options.configuration,
            options.section,
            options.c2ctemplate_style
        )

    theme_config = Config.get_theme_config_by_code(options.theme)
    if theme_config is None:
        parser.error('Specified theme not found in configuration.')

    if options.sql_file is None:
        create_theme_tables_(
            theme_config,
            tables_only=options.tables_only
        )
    else:
        with open(options.sql_file, 'w') as sql_file:
            create_theme_tables_(
                theme_config,
                tables_only=options.tables_only,
                sql_file=sql_file
            )



# -*- coding: utf-8 -*-
from pyramid_oereb.standard.xtf_import.util import parse_multilingual_text


class ViewService(object):

    TAG_REFERENCE_WMS = 'VerweisWMS'
    TAG_LEGEND = 'Legende'

    def __init__(self, session, model, legend_entry):
        self._session = session
        self._model = model
        self._legend_entry = legend_entry

    def parse(self, view_service):  # pragma: no cover
        reference_wms = parse_multilingual_text(view_service, self.TAG_REFERENCE_WMS)
        instance = self._model(
            id=view_service.attrib['TID'],
            reference_wms=reference_wms
        )
        self._session.add(instance)
        self._legend_entry.parse(view_service)



# -*- coding: utf-8 -*-
"""
This Package provides all models fitting to the standard database configuration. It is separated by the 17
mandatory topics the federation asks for.
"""


from sqlalchemy import Column, ForeignKey
from sqlalchemy import Boolean, String, Integer, Float, DateTime, Date
from geoalchemy2.types import Geometry as GeoAlchemyGeometry
from sqlalchemy.orm import relationship
from sqlalchemy_utils import JSONType


NAMING_CONVENTION = {
    "ix": 'ix_%(column_0_label)s',
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s"
}


def get_office(base, schema_name, pk_type):
    """
    Factory to produce a generic office model.

    Args:
        base (sqlalchemy.orm.decl_api.DeclarativeMeta): The SQLAlchemy base which is assigned to the models.
        schema_name (str): The name of the database schema where this models belong to.
        pk_type (sqlalchemy.sql.type_api.TypeEngine): The type of the primary column. E.g.
            sqlalchemy.String or sqlalchemy.Integer or another one fitting the underlying DB
            needs
    Retruns:
        sqlalchemy.orm.decl_api.DeclarativeMeta: The generated office model.
    """

    class Office(base):
        """
        The bucket to fill in all the offices you need to reference from public law restriction, document,
        geometry.

        Attributes:
            id (str): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
            name (dict): The multilingual name of the office.
            office_at_web (dict): A web accessible url to a presentation of this office (multilingual).
            uid (str): The uid of this office from https
            line1 (str): The first address line for this office.
            line2 (str): The second address line for this office.
            street (str): The streets name of the offices address.
            number (str): The number on street.
            postal_code (int): The ZIP-code.
            city (str): The name of the city.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'office'
        id = Column(pk_type, primary_key=True, autoincrement=False)
        name = Column(JSONType, nullable=False)
        office_at_web = Column(JSONType, nullable=True)
        uid = Column(String(12), nullable=True)
        line1 = Column(String, nullable=True)
        line2 = Column(String, nullable=True)
        street = Column(String, nullable=True)
        number = Column(String, nullable=True)
        postal_code = Column(Integer, nullable=True)
        city = Column(String, nullable=True)

    return Office


def get_document(base, schema_name, pk_type, Office):
    """
    Factory to produce a generic document model.

    Args:
        base (sqlalchemy.orm.decl_api.DeclarativeMeta): The SQLAlchemy base which is assigned to the models.
        schema_name (str): The name of the database schema where this models belong to.
        pk_type (sqlalchemy.sql.type_api.TypeEngine): The type of the primary column. E.g.
            sqlalchemy.String or sqlalchemy.Integer or another one fitting the underlying DB
            needs
        Office (sqlalchemy.orm.decl_api.DeclarativeMeta): The office model.
    Retruns:
        sqlalchemy.orm.decl_api.DeclarativeMeta: The generated office model.
    """

    class Document(base):
        """
        THE DOCUMENT
        This represents the main document in the whole system.
        Attributes:
            id (int): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
            document_type (str): The document type. It must be "Rechtsvorschrift", "GesetzlicheGrundlage"
                or "Hinweis".
            index (int): An index used to sort the documents.
            law_status (str): The status switch if the document is legally approved or not.
            title (dict): The multilingual title or if existing the short title ot his document.
            office_id (str): The foreign key to the office which is in charge for this document.
            responsible_office (pyramid_oereb.standard.models.railways_project_planning_zones.Office):
                The dedicated relation to the office instance from database.
            published_from (datetime.date): The date from when the document should be available for
                publishing in extracts. This  directly affects the behaviour of extract
                generation.
            published_until (datetime.date): The date until when the document should be available for
                publishing on extracts. This  directly affects the behaviour of extract
                generation.
            text_at_web (dict): A multilingual link which leads to the documents content in the web.
            abbreviation (dict): The multilingual shortened version of the documents title.
            official_number (dict): The multilingual official number which uniquely identifies this document.
            only_in_municipality (int): The fosnr (=id bfs) of the municipality. If this is None it is assumed
                the document is  related to the whole canton or even the confederation.
            file (str): The document itself as a binary representation (PDF). It is string but
                BaseCode64 encoded.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'document'
        id = Column(pk_type, primary_key=True, autoincrement=False)
        document_type = Column(String, nullable=False)
        index = Column(Integer, nullable=False)
        law_status = Column(String, nullable=False)
        title = Column(JSONType, nullable=False)
        office_id = Column(
            ForeignKey(Office.id),
            nullable=False
        )
        responsible_office = relationship(Office)
        published_from = Column(Date, nullable=False)
        published_until = Column(Date, nullable=True)
        text_at_web = Column(JSONType, nullable=True)
        abbreviation = Column(JSONType, nullable=True)
        official_number = Column(JSONType, nullable=True)
        only_in_municipality = Column(Integer, nullable=True)
        file = Column(String, nullable=True)

    return Document


def get_availablility(base, schema_name, pk_type):
    """
    Factory to produce a generic availability model.

    Args:
        base (sqlalchemy.orm.decl_api.DeclarativeMeta): The SQLAlchemy base which is assigned to the models.
        schema_name (str): The name of the database schema where this models belong to.
        pk_type (sqlalchemy.sql.type_api.TypeEngine): The type of the primary column. E.g.
            sqlalchemy.String or sqlalchemy.Integer or another one fitting the underlying DB
            needs
    Retruns:
        sqlalchemy.orm.decl_api.DeclarativeMeta: The generated office model.
    """

    class Availability(base):
        """
        A simple bucket for achieving a switch per municipality. Here you can configure via the
        imported data if a public law restriction is available or not. You need to fill it with
        the data you provided in the app schemas municipality table (fosnr).

        Attributes:
            fosnr (int): The identifier of the municipality in your system (id_bfs = fosnr)
            available (bool): The switch field to configure if this plr is available for the
                municipality or not.  This field has direct influence on the applications
                behaviour. See documentation for more info.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'availability'
        fosnr = Column(pk_type, primary_key=True, autoincrement=False)
        available = Column(Boolean, nullable=False, default=False)

    return Availability


def get_data_integration(base, schema_name, pk_type, Office):
    """
    Factory to produce a generic data integration model.

    Args:
        base (sqlalchemy.orm.decl_api.DeclarativeMeta): The SQLAlchemy base which is assigned to the models.
        schema_name (str): The name of the database schema where this models belong to.
        pk_type (sqlalchemy.sql.type_api.TypeEngine): The type of the primary column. E.g.
            sqlalchemy.String or sqlalchemy.Integer or another one fitting the underlying DB
            needs
        Office (sqlalchemy.orm.decl_api.DeclarativeMeta): The office model.
    Retruns:
        sqlalchemy.orm.decl_api.DeclarativeMeta: The generated office model.
    """

    class DataIntegration(base):
        """
        The bucket to fill in the date when this whole schema was updated. It has a relation to the
        office to be able to find out who was the delivering instance.

        Attributes:
            id (str): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
            date (datetime.date): The date when this data set was delivered.
            office_id (str): A foreign key which points to the actual office instance.
            office (pyramid_oereb.standard.models.airports_building_lines.Office):
                The actual office instance which the id points to.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'data_integration'
        id = Column(pk_type, primary_key=True, autoincrement=False)
        date = Column(DateTime, nullable=False)
        office_id = Column(ForeignKey(Office.id), nullable=False)
        office = relationship(Office)
        checksum = Column(String, nullable=True)

    return DataIntegration


def get_view_service(base, schema_name, pk_type):
    """
    Factory to produce a generic view service model.

    Args:
        base (sqlalchemy.orm.decl_api.DeclarativeMeta): The SQLAlchemy base which is assigned to the models.
        schema_name (str): The name of the database schema where this models belong to.
        pk_type (sqlalchemy.sql.type_api.TypeEngine): The type of the primary column. E.g.
            sqlalchemy.String or sqlalchemy.Integer or another one fitting the underlying DB
            needs
    Retruns:
        sqlalchemy.orm.decl_api.DeclarativeMeta: The generated office model.
    """

    class ViewService(base):
        """
        A view service aka WM(T)S which can deliver a cartographic representation via web.

        Attributes:
            id (str): The identifier. This is used in the database only and must not be set
                manually. If you  don't like it - don't care about.
            reference_wms (dict of str): The actual url which leads to the desired cartographic
                representation (multilingual).
            layer_index (int): Index for sorting the layering of the view services for a theme
            layer_opacity (float): Opacity of a view service
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'view_service'
        id = Column(pk_type, primary_key=True, autoincrement=False)
        reference_wms = Column(JSONType, nullable=False)
        layer_index = Column(Integer, nullable=False)
        layer_opacity = Column(Float, nullable=False)

    return ViewService


def get_legend_entry(base, schema_name, pk_type, ViewService):
    """
    Factory to produce a generic legend entry model.

    Args:
        base (sqlalchemy.orm.decl_api.DeclarativeMeta): The SQLAlchemy base which is assigned to the models.
        schema_name (str): The name of the database schema where this models belong to.
        pk_type (sqlalchemy.sql.type_api.TypeEngine): The type of the primary column. E.g.
            sqlalchemy.String or sqlalchemy.Integer or another one fitting the underlying DB
            needs
        ViewService (sqlalchemy.orm.decl_api.DeclarativeMeta): The view service model.
    Retruns:
        sqlalchemy.orm.decl_api.DeclarativeMeta: The generated office model.
    """

    class LegendEntry(base):
        """
        A class based legend system which is directly related to :class:`ViewService`.
        Attributes:
            id (str): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
            symbol (str): An image with represents the legend entry. This can be png or svg. It is string
                but BaseCode64  encoded.
            legend_text (dict): Multilingual text to describe this legend entry.
            type_code (str): Type code of the public law restriction which is represented by this legend
                entry.
            type_code_list (str): List of all public law restrictions which are described through this
                legend  entry.
            theme (str): Theme code to link this legend entry with the correct public law restriction.
            sub_theme (str): Sub-theme code to link this legend entry with the sub-theme of the public
                law restriction if there is a sub-theme.
            view_service_id (str): The foreign key to the view service this legend entry is related to.
            view_service (ViewService):
                The dedicated relation to the view service instance from database.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'legend_entry'
        id = Column(pk_type, primary_key=True, autoincrement=False)
        symbol = Column(String, nullable=False)
        legend_text = Column(JSONType, nullable=False)
        type_code = Column(String(40), nullable=False)
        type_code_list = Column(String, nullable=False)
        theme = Column(String, nullable=False)
        sub_theme = Column(String, nullable=True)
        view_service_id = Column(
            ForeignKey(ViewService.id),
            nullable=False
        )
        view_service = relationship(ViewService, backref='legends')

    return LegendEntry


def get_public_law_restriction(base, schema_name, pk_type, Office, ViewService, LegendEntry):
    """
    Factory to produce a generic public law restriction model.

    Args:
        base (sqlalchemy.orm.decl_api.DeclarativeMeta): The SQLAlchemy base which is assigned to the models.
        schema_name (str): The name of the database schema where this models belong to.
        pk_type (sqlalchemy.sql.type_api.TypeEngine): The type of the primary column. E.g.
            sqlalchemy.String or sqlalchemy.Integer or another one fitting the underlying DB
            needs
        Office (sqlalchemy.orm.decl_api.DeclarativeMeta): The office model.
        ViewService (sqlalchemy.orm.decl_api.DeclarativeMeta): The view service model.
        LegendEntry (sqlalchemy.orm.decl_api.DeclarativeMeta): The legend entry model.
    Retruns:
        sqlalchemy.orm.decl_api.DeclarativeMeta: The generated office model.
    """

    class PublicLawRestriction(base):
        """
        The container where you can fill in all your public law restrictions to the topic.
        Attributes:
            id (str): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
            law_status (str): The status switch if the document is legally approved or not.
            published_from (datetime.date): The date when the document should be available for
                publishing on extracts. This  directly affects the behaviour of extract
                generation.
            published_until (datetime.date): The date starting from which the document should not be
                published anymore on extracts. This directly affects the behaviour of extract generation.
            view_service_id (str): The foreign key to the view service this public law restriction is
                related to.
            view_service (ViewService):
                The dedicated relation to the view service instance from database.
            office_id (str): The foreign key to the office which is responsible to this public law
                restriction.
            responsible_office (Office):
                The dedicated relation to the office instance from database.
            legend_entry_id (str): The foreign key to the legend entry this public law restriction is
                related to.
            legend_entry (pyramid_oereb.standard.models.airports_building_lines.LegendEntry):
                The dedicated relation to the legend entry instance from database.
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'public_law_restriction'
        id = Column(pk_type, primary_key=True, autoincrement=False)
        law_status = Column(String, nullable=False)
        published_from = Column(Date, nullable=False)
        published_until = Column(Date, nullable=True)
        view_service_id = Column(
            ForeignKey(ViewService.id),
            nullable=False
        )
        view_service = relationship(
            ViewService,
            backref='public_law_restrictions'
        )
        office_id = Column(
            ForeignKey(Office.id),
            nullable=False
        )
        responsible_office = relationship(Office)
        legend_entry_id = Column(
            ForeignKey(LegendEntry.id),
            nullable=False
        )
        legend_entry = relationship('LegendEntry', backref='public_law_restrictions')

    return PublicLawRestriction


def get_geometry(base, schema_name, pk_type, geometry_type, srid, PublicLawRestriction):
    """
    Factory to produce a generic geometry model.

    Args:
        base (sqlalchemy.orm.decl_api.DeclarativeMeta): The SQLAlchemy base which is assigned to the models.
        schema_name (str): The name of the database schema where this models belong to.
        pk_type (sqlalchemy.sql.type_api.TypeEngine): The type of the primary column. E.g.
            sqlalchemy.String or sqlalchemy.Integer or another one fitting the underlying DB
            needs
        geometry_type (str): The geoalchemy geometry type defined as well known string.
        srid (int): The SRID defining the projection of the geometries stored in standard db schema.
        PublicLawRestriction (sqlalchemy.orm.decl_api.DeclarativeMeta): The public law restriction model.
    Retruns:
        sqlalchemy.orm.decl_api.DeclarativeMeta: The generated office model.
    """

    class Geometry(base):
        """
        The dedicated model for all geometries in relation to their public law restriction.
        Attributes:
            id (str): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
            law_status (str): The status switch if the document is legally approved or not.
            published_from (datetime.date): The date when the geometry should be available for
                publishing on extracts. This directly affects the behaviour of extract
                generation.
            published_until (datetime.date): The date from when the geometry should not be available
                anymore for publishing on extracts. This directly affects the behaviour of extract
                generation.
            geo_metadata (str): A link to the metadata which this geometry is based on which delivers
                machine  readable response format (XML).
            public_law_restriction_id (str): The foreign key to the public law restriction this geometry
                is  related to.
            public_law_restriction (PublicLawRestriction): The dedicated relation to the public law
                restriction instance from database.
            geom (geoalchemy2.types.Geometry): The geometry it's self. For type information see
                geoalchemy docs (https://geoalchemy-2.readthedocs.io/en/0.4.2/types.html) dependent on the
                configured type.  This concrete one is LINESTRING
        """
        __table_args__ = {'schema': schema_name}
        __tablename__ = 'geometry'
        id = Column(pk_type, primary_key=True, autoincrement=False)
        law_status = Column(String, nullable=False)
        published_from = Column(Date, nullable=False)
        published_until = Column(Date, nullable=True)
        geo_metadata = Column(String, nullable=True)
        geom = Column(GeoAlchemyGeometry(geometry_type, srid=srid), nullable=False)
        public_law_restriction_id = Column(
            ForeignKey(PublicLawRestriction.id),
            nullable=False
        )
        public_law_restriction = relationship(
            PublicLawRestriction,
            backref='geometries'
        )

    return Geometry


def get_public_law_restriction_document(base, schema_name, pk_type, PublicLawRestriction, Document):
    """
    Factory to produce a generic public law restriction document model.

    Args:
        base (sqlalchemy.orm.decl_api.DeclarativeMeta): The SQLAlchemy base which is assigned to the models.
        schema_name (str): The name of the database schema where this models belong to.
        pk_type (sqlalchemy.sql.type_api.TypeEngine): The type of the primary column. E.g.
            sqlalchemy.String or sqlalchemy.Integer or another one fitting the underlying DB
            needs
        PublicLawRestriction (sqlalchemy.orm.decl_api.DeclarativeMeta): The public law restriction model.
        Document (sqlalchemy.orm.decl_api.DeclarativeMeta): The document model.
    Retruns:
        sqlalchemy.orm.decl_api.DeclarativeMeta: The generated office model.
    """

    class PublicLawRestrictionDocument(base):
        """
        Meta bucket (join table) for the relationship between public law restrictions and documents.
        Attributes:
            id (str): The identifier. This is used in the database only and must not be set manually. If
                you  don't like it - don't care about.
            public_law_restriction_id (str): The foreign key to the public law restriction which has
                relation to  a document.
            document_id (str): The foreign key to the document which has relation to the public law
                restriction.
            plr (PublicLawRestriction):
                The dedicated relation to the public law restriction instance from database.
            document (Document):
                The dedicated relation to the document instance from database.
        """
        __tablename__ = 'public_law_restriction_document'
        __table_args__ = {'schema': schema_name}
        id = Column(pk_type, primary_key=True, autoincrement=False)
        public_law_restriction_id = Column(
            ForeignKey(PublicLawRestriction.id),
            nullable=False
        )
        document_id = Column(
            ForeignKey(Document.id),
            nullable=False
        )
        plr = relationship(
            PublicLawRestriction,
            backref='legal_provisions'
        )
        document = relationship(
            Document
        )

    return PublicLawRestrictionDocument



# -*- coding: utf-8 -*-
from pyramid_oereb.lib.sources import BaseDatabaseSource
from pyramid_oereb.lib.sources.disclaimer import DisclaimerBaseSource
from pyramid_oereb.contrib.interlis.interlis_2_3_utils import from_multilingual_text_to_dict


class DatabaseSource(BaseDatabaseSource, DisclaimerBaseSource):

    def read(self, params):
        """
        The read method to access the standard database structure. It uses SQL-Alchemy for querying. It does
        not accept any parameters nor it applies any filter on the database query. It simply loads all
        content from the configured model.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
        """
        session = self._adapter_.get_session(self._key_)
        try:
            results = session.query(self._model_).all()

            self.records = list()
            for result in results:
                self.records.append(self._record_class_(
                    from_multilingual_text_to_dict(
                        de=result.title_de,
                        fr=result.title_fr,
                        it=result.title_it,
                        rm=result.title_rm,
                        en=result.title_en),
                    from_multilingual_text_to_dict(
                        de=result.content_de,
                        fr=result.content_fr,
                        it=result.content_it,
                        rm=result.content_rm,
                        en=result.content_en)
                ))
        finally:
            session.close()



# -*- coding: utf-8 -*-

import re

from sqlalchemy import create_engine
from sqlalchemy.schema import CreateTable


def convert_camel_case_to_snake_case(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def convert_camel_case_to_text_form(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1 \2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1 \2', s1)


def tables(base):
    return base.metadata.sorted_tables


def create_schema_sql(schema_name):
    return 'CREATE SCHEMA IF NOT EXISTS {0};'.format(schema_name)


def create_tables_sql(db_connection, tables):
    """
    Args:
        db_connection (str): The db connection string.
        tables (list of sqlalchemy.schema.Table): The table objects from sqlalchemy.
    Returns:
        a string with the sql statement used to create the tables
    """
    sqls = []
    engine = create_engine(db_connection, echo=True)
    for table in tables:
        sqls.append(
            '{};\n'.format(
                str(CreateTable(table).compile(engine)).replace('DATETIME', 'timestamp')
            )
        )
    return ''.join(sqls)


def create_sql(schema_name, db_connection, tables):
    sqls = [
        create_schema_sql(schema_name),
        create_tables_sql(db_connection, tables)
    ]
    return '\n'.join(sqls)


def drop_sql(schema_name):
    return 'DROP SCHEMA IF EXISTS {0} CASCADE;'.format(schema_name)


def execute_sql(db_connection, sql):
    connection = create_engine(db_connection, echo=True).connect()
    try:
        connection.execute(sql)
    finally:
        connection.close()



# -*- coding: utf-8 -*-
from pyramid_oereb.lib.sources import Base
from pyramid_oereb.lib.records.real_estate_type import RealEstateTypeRecord


class RealEstateTypeBaseSource(Base):
    """
    Base class for real estate type values source.
    Attributes:
        records (list of pyramid_oereb.lib.records.real_estate_type.RealEstateTypeRecord): List of real estate
        type records.
    """
    _record_class_ = RealEstateTypeRecord

    def read(self, params=None):
        """
        Every real estate type source has to implement a read method. This method must accept no parameters.
        Because it should deliver all items available.
        If you want adapt to your own source for real estate type labels, this is the point where to hook in.
        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
        """
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
from pyramid_oereb.standard.xtf_import.util import parse_string, parse_multilingual_text


class Office(object):

    TAG_NAME = 'Name'
    TAG_OFFICE_AT_WEB = 'AmtImWeb'
    TAG_UID = 'UID'

    def __init__(self, session, model):
        self._session = session
        self._model = model

    def parse(self, office):  # pragma: no cover
        instance = self._model(
            id=office.attrib['TID'],
            name=parse_multilingual_text(office, self.TAG_NAME),
            office_at_web=parse_multilingual_text(office, self.TAG_OFFICE_AT_WEB),
            uid=parse_string(office, self.TAG_UID)
        )
        self._session.add(instance)



# -*- coding: utf-8 -*-
import logging
import importlib
import binascii

from geoalchemy2.shape import to_shape, from_shape
from shapely.geometry import Point, LineString, Polygon, MultiPoint, MultiLineString, MultiPolygon, \
    GeometryCollection
from sqlalchemy import text, or_

from pyramid_oereb import Config
from pyramid_oereb.lib import b64
from pyramid_oereb.lib.records.availability import AvailabilityRecord
from pyramid_oereb.lib.records.image import ImageRecord
from pyramid_oereb.lib.records.plr import EmptyPlrRecord
from pyramid_oereb.lib.sources import BaseDatabaseSource
from pyramid_oereb.lib.sources.plr import PlrBaseSource
from pyramid_oereb.contrib.interlis.interlis_2_3_utils import from_multilingual_text_to_dict
from pyramid_oereb.contrib.interlis.interlis_2_3_utils import from_multilingual_uri_to_dict

log = logging.getLogger(__name__)


class StandardThemeConfigParser(object):

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @property
    def model_factory_path(self):
        return self.kwargs.get('source').get('params').get('model_factory')

    @property
    def schema_name(self):
        return self.kwargs.get('source').get('params').get('schema_name')

    @property
    def geometry_type(self):
        return self.kwargs.get('geometry_type')

    @property
    def srid(self):
        return Config._config.get('srid')

    @property
    def db_connection(self):
        return self.kwargs.get('source').get('params').get('db_connection')

    def get_model_factory(self):
        module_elements = Config.extract_module_function(
            self.model_factory_path
        )
        return getattr(
            importlib.import_module(module_elements['module_path']),
            module_elements['function_name']
        )

    def get_models(self):
        model_factory = self.get_model_factory()
        return model_factory(
            self.schema_name,
            self.geometry_type,
            self.srid,
            self.db_connection
        )


def parse_multiple_standard_themes(config):
    """
    Args:
        config (pyramid_oereb.standard.config.Config): The config object of the application.
    Returns:
        dict containing the factory_model for every plr code.
    """
    plrs = config._config.get('plrs')
    themes = {}
    for plr in plrs:
        config_parser = StandardThemeConfigParser(**plr)
        themes[plr['code']] = config_parser.get_models()
    return themes


class DatabaseSource(BaseDatabaseSource, PlrBaseSource):
    def __init__(self, **kwargs):
        """
        Keyword Arguments:
            name (str): The name. You are free to choose one.
            code (str): The official code. Regarding to the federal specifications.
            geometry_type (str): The geometry type. Possible are: POINT, POLYGON, LINESTRING,
                GEOMETRYCOLLECTION
            thresholds (dict): The configuration of limits and units used for processing.
            text (dict of str): The speaking title. It must be a dictionary containing language (as
                configured) as key and text as value.
            language (str): The language this public law restriction is originally shipped with.
            federal (bool): Switch if it is a federal topic. This will be taken into account in processing
                steps.
            source (dict): The configuration dictionary of the public law restriction
            hooks (dict of str): The hook methods: get_symbol, get_symbol_ref. They have to be provided as
                dotted string for further use with dotted name resolver of pyramid package.
            law_status (dict of str): The configuration dictionary of the law status. It consists of
                the code and text which must be a dictionary containing language (as configured)
                as key and text as value.
        """
        config_parser = StandardThemeConfigParser(**kwargs)
        models = config_parser.get_models()
        bds_kwargs = {
            'model': models.Geometry,
            'db_connection': kwargs.get('source').get('params').get('db_connection')
        }

        BaseDatabaseSource.__init__(self, **bds_kwargs)
        PlrBaseSource.__init__(self, **kwargs)

        self.legend_entry_model = models.LegendEntry
        availability_model = models.Availability

        self.availabilities = []
        self.datasource = []

        session = self._adapter_.get_session(self._key_)

        try:

            availabilities_from_db = session.query(availability_model).all()
            for availability in availabilities_from_db:
                self.availabilities.append(
                    AvailabilityRecord(availability.fosnr, available=availability.available)
                )

        finally:
            session.close()

    def from_db_to_legend_entry_record(self, legend_entry_from_db):
        theme = Config.get_theme_by_code_sub_code(legend_entry_from_db.theme, legend_entry_from_db.sub_theme)
        legend_entry_record = self._legend_entry_record_class(
            ImageRecord(
                b64.decode(
                    binascii.b2a_base64(legend_entry_from_db.symbol).decode('ascii')
                )
            ),
            from_multilingual_text_to_dict(
                de=legend_entry_from_db.legend_text_de,
                fr=legend_entry_from_db.legend_text_fr,
                it=legend_entry_from_db.legend_text_it,
                rm=legend_entry_from_db.legend_text_rm,
                en=legend_entry_from_db.legend_text_en
            ),
            legend_entry_from_db.type_code,
            legend_entry_from_db.type_code_list,
            theme,
            view_service_id=legend_entry_from_db.view_service_id,
        )
        return legend_entry_record

    def from_db_to_legend_entry_records(self, legend_entries_from_db, sub_theme=None):
        legend_entry_records = []
        if sub_theme is not None:
            for legend_entry_from_db in legend_entries_from_db:
                if legend_entry_from_db.sub_theme == sub_theme:
                    self.from_db_to_legend_entry_record(legend_entry_from_db)
        else:
            for legend_entry_from_db in legend_entries_from_db:
                legend_entry_records.append(
                    self.from_db_to_legend_entry_record(legend_entry_from_db)
                )

        return legend_entry_records

    def from_db_to_view_service_record(self, view_service_from_db, legend_entry_records, theme):
        multilingual_uri = from_multilingual_uri_to_dict(view_service_from_db.multilingual_uri)
        layer_index, layer_opacity = Config.get_index_and_opacity_of_view_service(multilingual_uri)
        view_service_record = self._view_service_record_class(
            multilingual_uri,
            layer_index,
            layer_opacity,
            legends=legend_entry_records
        )
        return view_service_record

    def unwrap_multi_geometry_(self, law_status, published_from, published_until,
                               multi_geom, geo_metadata):
        unwrapped = []
        for geom in multi_geom.geoms:
            unwrapped.append(
                self._geometry_record_class(
                    law_status,
                    published_from,
                    published_until,
                    geom,
                    geo_metadata
                )
            )
        return unwrapped

    def unwrap_geometry_collection_(self, law_status, published_from, published_until,
                                    collection, geo_metadata):
        unwrapped = []
        if len(collection.geoms) > 1:
            raise AttributeError(u'There was more than one element in the GeometryCollection. '
                                 u'This is not supported!')
        for geom in collection.geoms:
            unwrapped.extend(
                self.create_geometry_records_(
                    law_status,
                    published_from,
                    published_until,
                    geom,
                    geo_metadata
                )
            )
        return unwrapped

    def create_geometry_records_(self, law_status, published_from, published_until,
                                 geom, geo_metadata):
        geometry_records = []

        # Process single geometries
        if isinstance(geom, (Point, LineString, Polygon)):
            geometry_records.append(
                self._geometry_record_class(
                    law_status,
                    published_from,
                    published_until,
                    geom,
                    geo_metadata
                )
            )

        # Process multi geometries
        elif isinstance(geom, (MultiPoint, MultiLineString, MultiPolygon)):
            geometry_records.extend(self.unwrap_multi_geometry_(
                law_status,
                published_from,
                published_until,
                geom,
                geo_metadata
            ))

        # Process geometry collections
        elif isinstance(geom, GeometryCollection):
            geometry_records.extend(self.unwrap_geometry_collection_(
                law_status,
                published_from,
                published_until,
                geom,
                geo_metadata
            ))

        else:
            log.warning('Received geometry with unsupported type: {0}. Skipped geometry.'.format(
                geom.type
            ))

        return geometry_records

    def from_db_to_geometry_records(self, geometries_from_db):
        geometry_records = []
        for geometry_from_db in geometries_from_db:
            # Create law status record
            law_status = Config.get_law_status_by_data_code(
                    self._plr_info.get('code'),
                    geometry_from_db.law_status
                )

            if geometry_from_db.point is not None:
                geom = geometry_from_db.point
            elif geometry_from_db.line is not None:
                geom = geometry_from_db.line
            elif geometry_from_db.surface is not None:
                geom = geometry_from_db.surface

            # Create geometry records
            geometry_records.extend(self.create_geometry_records_(
                law_status,
                geometry_from_db.published_from,
                geometry_from_db.published_until,
                to_shape(geom),
                geometry_from_db.geo_metadata
            ))

        return geometry_records

    def from_db_to_office_record(self, offices_from_db):
        office_record = self._office_record_class(
            from_multilingual_text_to_dict(
                de=offices_from_db.name_de,
                fr=offices_from_db.name_fr,
                it=offices_from_db.name_it,
                rm=offices_from_db.name_rm,
                en=offices_from_db.name_en,
            ),
            offices_from_db.uid,
            from_multilingual_uri_to_dict(offices_from_db.multilingual_uri),
            offices_from_db.line1,
            offices_from_db.line2,
            offices_from_db.street,
            offices_from_db.number,
            offices_from_db.postal_code,
            offices_from_db.city
        )

        return office_record

    def from_db_to_document_records(self, documents_from_db):
        document_records = []
        for document in documents_from_db:
            office_record = self.from_db_to_office_record(document.responsible_office)
            law_status = Config.get_law_status_by_data_code(
                self._plr_info.get('code'),
                document.law_status
            )
            document_records.append(self._documents_record_class(
                document_type=Config.get_document_type_by_data_code(
                    self._plr_info.get('code'),
                    document.document_type
                ),
                index=document.index,
                law_status=law_status,
                title=from_multilingual_text_to_dict(
                        de=document.title_de,
                        fr=document.title_fr,
                        it=document.title_it,
                        rm=document.title_rm,
                        en=document.title_en),
                responsible_office=office_record,
                published_from=document.published_from,
                published_until=document.published_until,
                text_at_web=from_multilingual_uri_to_dict(document.multilingual_uri),
                abbreviation=from_multilingual_text_to_dict(
                        de=document.abbreviation_de,
                        fr=document.abbreviation_fr,
                        it=document.abbreviation_it,
                        rm=document.abbreviation_rm,
                        en=document.abbreviation_en),
                official_number=from_multilingual_text_to_dict(
                        de=document.official_number_de,
                        fr=document.official_number_fr,
                        it=document.official_number_it,
                        rm=document.official_number_rm,
                        en=document.official_number_en),
                only_in_municipality=document.only_in_municipality,
                article_numbers=None,
                file=None
            ))
        return document_records

    def from_db_to_plr_record(self, params, public_law_restriction_from_db, legend_entries_from_db):
        thresholds = self._plr_info.get('thresholds')
        min_length = thresholds.get('length').get('limit')
        length_unit = thresholds.get('length').get('unit')
        min_area = thresholds.get('area').get('limit')
        area_unit = thresholds.get('area').get('unit')
        legend_entry_record = self.from_db_to_legend_entry_record(
            public_law_restriction_from_db.legend_entry
        )
        legend_entry_records = self.from_db_to_legend_entry_records(
            legend_entries_from_db,
            legend_entry_record.theme.sub_code
        )
        symbol = ImageRecord(
            b64.decode(
                binascii.b2a_base64(public_law_restriction_from_db.legend_entry.symbol).decode('ascii')
            )
        )
        view_service_record = self.from_db_to_view_service_record(
            public_law_restriction_from_db.view_service,
            legend_entry_records,
            self._plr_info.get('code')
        )
        theme = Config.get_theme_by_code_sub_code(
            public_law_restriction_from_db.legend_entry.theme,
            public_law_restriction_from_db.legend_entry.sub_theme)
        if theme.document_records is None:
            document_records = self.get_document_records(params, public_law_restriction_from_db)
        else:
            document_records = theme.document_records +\
                self.get_document_records(params, public_law_restriction_from_db)
        geometry_records = self.from_db_to_geometry_records(public_law_restriction_from_db.geometries)
        law_status = Config.get_law_status_by_data_code(
            self._plr_info.get('code'),
            public_law_restriction_from_db.law_status
        )
        sub_theme_legend_entry_record = \
            legend_entry_record.theme if legend_entry_record.theme.sub_code else None
        plr_record = self._plr_record_class(
            Config.get_theme_by_code_sub_code(legend_entry_record.theme.code),
            legend_entry_record,
            law_status,
            public_law_restriction_from_db.published_from,
            public_law_restriction_from_db.published_until,
            self.from_db_to_office_record(public_law_restriction_from_db.responsible_office),
            symbol,
            view_service_record,
            geometry_records,
            sub_theme=sub_theme_legend_entry_record,
            type_code=public_law_restriction_from_db.legend_entry.type_code,
            type_code_list=public_law_restriction_from_db.legend_entry.type_code_list,
            documents=document_records,
            min_area=min_area,
            min_length=min_length,
            area_unit=area_unit,
            length_unit=length_unit,
            view_service_id=public_law_restriction_from_db.view_service.t_id
        )

        return plr_record

    def get_document_records(self, params, public_law_restriction_from_db):
        documents_from_db = []
        if not hasattr(public_law_restriction_from_db, 'legal_provisions'):
            raise AttributeError('The public_law_restriction implementation of type {} has no '
                                 'legal_provisions attribute. Check the model implementation.'
                                 .format(type(public_law_restriction_from_db)))
        for legal_provision in public_law_restriction_from_db.legal_provisions:
            documents_from_db.append(legal_provision.document)
        document_records = self.from_db_to_document_records(documents_from_db)
        return document_records

    @staticmethod
    def extract_geometry_collection_db(db_path, real_estate_geometry):
        """
        Decides the geometry collection cases of geometric filter operations when the database contains multi
        geometries but the passed geometry does not.
        The multi geometry will be extracted to it's sub parts for operation.

        Args:
            db_path (str): The point separated string of schema_name.table_name.column_name from
                which we can construct a correct SQL statement.
            real_estate_geometry (shapely.geometry.base.BaseGeometry): The shapely geometry
                representation which is used for comparison.

        Returns:
            sqlalchemy.sql.elements.BooleanClauseList: The clause element.

        Raises:
            HTTPBadRequest
        """
        srid = Config.get('srid')
        sql_text_point = 'ST_Intersects(ST_CollectionExtract({0}, 1), ST_GeomFromText(\'{1}\', {2}))'.format(
            db_path,
            real_estate_geometry.wkt,
            srid
        )
        sql_text_line = 'ST_Intersects(ST_CollectionExtract({0}, 2), ST_GeomFromText(\'{1}\', {2}))'.format(
            db_path,
            real_estate_geometry.wkt,
            srid
        )
        sql_text_polygon = 'ST_Intersects(ST_CollectionExtract({0}, 3), ' \
                           'ST_GeomFromText(\'{1}\', {2}))'.format(
                                db_path,
                                real_estate_geometry.wkt,
                                srid
                            )
        clause_blocks = [
            text(sql_text_point),
            text(sql_text_line),
            text(sql_text_polygon)
        ]
        return or_(*clause_blocks)

    def collect_related_geometries_by_real_estate(self, session, real_estate):
        """
        Extracts all geometries in the topic which have spatial relation with the passed real estate

        Args:
            session (sqlalchemy.orm.Session): The requested clean session instance ready for use
            real_estate (pyramid_oereb.lib.records.real_estate.RealEstateRecord): The real
                estate in its record representation.

        Returns:
            list: The result of the related geometries unique by the public law restriction id
        """

        return session.query(self._model_).filter(
                 or_(
                  self._model_.point.ST_Intersects(from_shape(real_estate.limit, srid=Config.get('srid'))),
                  self._model_.line.ST_Intersects(from_shape(real_estate.limit, srid=Config.get('srid'))),
                  self._model_.surface.ST_Intersects(from_shape(real_estate.limit, srid=Config.get('srid')))
                 )).distinct(self._model_.public_law_restriction_id).all()

    def collect_legend_entries_by_bbox(self, session, bbox):
        """
        Extracts all legend entries in the topic which have spatial relation with the passed bounding box of
        visible extent.

        Args:
            session (sqlalchemy.orm.Session): The requested clean session instance ready for use
            bbox (shapely.geometry.base.BaseGeometry): The bbox to search the records.

        Returns:
            list: The result of the related geometries unique by the public law restriction id
        """
        distinct_legend_entry_ids = []
        geometries = session.query(self._model_).filter(
                      or_(
                       self._model_.point.ST_Intersects(from_shape(bbox, srid=Config.get('srid'))),
                       self._model_.line.ST_Intersects(from_shape(bbox, srid=Config.get('srid'))),
                       self._model_.surface.ST_Intersects(from_shape(bbox, srid=Config.get('srid')))
                      )).distinct(self._model_.public_law_restriction_id).all()

        for geometry in geometries:
            if geometry.public_law_restriction.legend_entry_id not in distinct_legend_entry_ids:
                distinct_legend_entry_ids.append(geometry.public_law_restriction.legend_entry_id)

        return session.query(self.legend_entry_model).filter(
            self.legend_entry_model.t_id.in_((distinct_legend_entry_ids))).all()

    def read(self, params, real_estate, bbox):
        """
        The read point which creates a extract, depending on a passed real estate.

        Args:
            params (pyramid_oereb.views.webservice.Parameter): The parameters of the extract request.
            real_estate (pyramid_oereb.lib.records.real_estate.RealEstateRecord): The real
                estate in its record representation.
            bbox (shapely.geometry.base.BaseGeometry): The bbox to search the records.
            position (int or None): relative position of the plr (within a list of plrs)
        """

        # Check if the plr is marked as available
        if self._is_available(real_estate):
            session = self._adapter_.get_session(self._key_)
            try:
                if session.query(self._model_).count() == 0:
                    # We can stop here already because there are no items in the database
                    self.records = [EmptyPlrRecord(
                            Config.get_theme_by_code_sub_code(self._plr_info['code'])
                        )]
                else:
                    # We need to investigate more in detail

                    # Try to find geometries which have spatial relation with real estate
                    geometry_results = self.collect_related_geometries_by_real_estate(
                        session, real_estate
                    )
                    if len(geometry_results) == 0:
                        # We checked if there are spatially related elements in database. But there is none.
                        # So we can stop here.
                        self.records = [EmptyPlrRecord(
                            Config.get_theme_by_code_sub_code(self._plr_info['code'])
                        )]
                    else:
                        # We found spatially related elements. This means we need to extract the actual plr
                        # information related to the found geometries.
                        self.records = []
                        legend_entries_from_db = self.collect_legend_entries_by_bbox(session, bbox)
                        for geometry_result in geometry_results:
                            self.records.append(
                                self.from_db_to_plr_record(
                                    params,
                                    geometry_result.public_law_restriction,
                                    legend_entries_from_db
                                )
                            )

            finally:
                session.close()

        # Add empty record if topic is not available
        else:
            self.records = [EmptyPlrRecord(self._theme_record, has_data=False)]

    def _is_available(self, real_estate):
        """
        Checks if the topic is available for the specified real estate.

        Args:
            real_estate (pyramid_oereb.lib.records.real_estate.RealEstateRecord): The real
                estate in its record representation.

        Returns:
             bool: True if the topic is available, false otherwise.
        """
        for availability in self.availabilities:
            if int(real_estate.fosnr) == int(availability.fosnr) and not availability.available:
                return False
        return True



# -*- coding: utf-8 -*-
import logging

from pyramid.httpexceptions import HTTPInternalServerError
from mako.lookup import TemplateLookup
from pyramid.path import AssetResolver

from pyramid.response import Response

from pyramid_oereb.lib.renderer import Base
from mako import exceptions

from pyramid_oereb.views.webservice import Parameter

log = logging.getLogger(__name__)


class Renderer(Base):

    def __init__(self, info):
        """
        Creates a new XML renderer instance for extract rendering.

        Args:
            info (pyramid.interfaces.IRendererInfo): Info object.
        """
        a = AssetResolver('pyramid_oereb')
        resolver = a.resolve('lib/renderer/extract/templates/xml')
        self.template_dir = resolver.abspath()
        self._gml_id = 0
        super(Renderer, self).__init__(info)

    def __call__(self, value, system):
        """
        Returns the XML encoded extract, according to the specification.

        Args:
            value (tuple): A tuple containing the generated extract record and the params
                dictionary.
            system (dict): The available system properties.

        Returns:
            str: The XML encoded extract.
        """
        self._request = self.get_request(system)
        response = self.get_response(system)
        if isinstance(response, Response) and response.content_type == response.default_content_type:
            response.content_type = 'application/xml'

        self._params_ = value[1]
        if not isinstance(self._params_, Parameter):
            raise TypeError('Missing parameter definition; Expected {0}, got {1} instead'.format(
                Parameter,
                self._params_.__class__
            ))
        if self._params_.language:
            self._language = str(self._params_.language).lower()

        extract = value[0]
        try:
            content = self._render(extract, self._params_)
            return content
        except ValueError as e:
            log.error('The extract can not be rendered. ValueError is {0}'.format(e))
            raise HTTPInternalServerError()
        except Exception:
            response.content_type = 'text/html'
            return exceptions.html_error_template().render()

    def _render(self, extract, params):
        templates = TemplateLookup(
            directories=[self.template_dir],
            output_encoding='utf-8',
            input_encoding='utf-8'
        )
        template = templates.get_template('extract.xml')
        content = template.render(**{
            'extract': extract,
            'params': params,
            'sort_by_localized_text': self.sort_by_localized_text,
            'localized': self.get_localized_text,
            'multilingual': self.get_multilingual_text,
            'get_localized_image': self.get_localized_image,
            'request': self._request,
            'get_symbol_ref': self.get_symbol_ref,
            'get_gml_id': self._get_gml_id,
            'date_format': '%Y-%m-%dT%H:%M:%S'
        })
        return content

    def _get_gml_id(self):
        """
        Returns the next GML ID.

        Returns:
             int: The next GML ID.

        """
        self._gml_id += 1
        return 'gml{0}'.format(self._gml_id)



# -*- coding: utf-8 -*-
from pyramid_oereb.lib.sources import BaseDatabaseSource
from pyramid_oereb.lib.sources.theme import ThemeBaseSource
from pyramid_oereb.contrib.interlis.interlis_2_3_utils import from_multilingual_text_to_dict
from pyramid_oereb.contrib.interlis.interlis_2_3_utils import from_multilingual_uri_to_dict
from pyramid_oereb import Config


class DatabaseSource(BaseDatabaseSource, ThemeBaseSource):

    def from_db_to_office_record(self, offices_from_db):
        office_record = self._office_record_class(
            offices_from_db.name,
            offices_from_db.uid,
            from_multilingual_uri_to_dict(offices_from_db.multilingual_uri),
            offices_from_db.line1,
            offices_from_db.line2,
            offices_from_db.street,
            offices_from_db.number,
            offices_from_db.postal_code,
            offices_from_db.city
        )

        return office_record

    def from_db_to_document_records(self, documents_from_db, article_numbers=None):
        document_records = []
        for i, document in enumerate(documents_from_db):
            office_record = self.from_db_to_office_record(document.responsible_office)

            article_nrs = article_numbers[i] if isinstance(article_numbers, list) else None
            law_status = Config.get_law_status_by_code(
                self._plr_info.get('code'),
                document.law_status
            )
            document_records.append(self._documents_record_class(
                document_type=document.document_type,
                index=document.index,
                law_status=law_status,
                title=from_multilingual_text_to_dict(
                        de=document.title_de,
                        fr=document.title_fr,
                        it=document.title_it,
                        rm=document.title_rm,
                        en=document.title_en),
                responsible_office=office_record,
                published_from=document.published_from,
                published_until=document.published_until,
                text_at_web=from_multilingual_uri_to_dict(document.multilingual_uri),
                abbreviation=from_multilingual_text_to_dict(
                        de=document.abbreviation_de,
                        fr=document.abbreviation_fr,
                        it=document.abbreviation_it,
                        rm=document.abbreviation_rm,
                        en=document.abbreviation_en),
                official_number=from_multilingual_text_to_dict(
                        de=document.official_number_de,
                        fr=document.official_number_fr,
                        it=document.official_number_it,
                        rm=document.official_number_rm,
                        en=document.official_number_en),
                only_in_municipality=document.only_in_municipality,
                article_numbers=article_nrs,
                file=document.file
            ))
        return document_records

    def get_document_records(self, theme_from_db):
        documents_from_db = []
        article_numbers = []
        if not hasattr(theme_from_db, 'legal_provisions'):
            raise AttributeError('The public_law_restriction implementation of type {} has no '
                                 'legal_provisions attribute. Check the model implementation.'
                                 .format(type(theme_from_db)))
        for legal_provision in theme_from_db.legal_provisions:
            documents_from_db.append(legal_provision.document)
            if len(legal_provision.article_numbers) == 0:
                article_numbers.append(None)
            else:
                article_nrs = []
                for article_number in legal_provision.article_numbers:
                    article_nrs.append(article_number.value)
                article_numbers.append(article_nrs)

        document_records = self.from_db_to_document_records(documents_from_db, article_numbers)
        return document_records

    def read(self):
        """
        Central method to read all theme entries.
        """
        session = self._adapter_.get_session(self._key_)
        try:
            results = session.query(self._model_).all()

            self.records = list()
            for result in results:
                self.records.append(self._record_class_(
                    result.code,
                    from_multilingual_text_to_dict(
                        de=result.title_de,
                        fr=result.title_fr,
                        it=result.title_it,
                        rm=result.title_rm,
                        en=result.title_en),
                    result.extract_index,
                    results.sub_code,
                    self.get_document_records(result)
                ))
        finally:
            session.close()
